"""
Executors
=========

This package contains the built-in executors for Chancy. Executors are
responsible for managing the execution of jobs after they've been retrieved
from a queue.

Chancy comes with two built-in executors:

- :class:`chancy.executors.process.ProcessExecutor`: This executor will execute jobs
    in a separate process. This is the default executor.
- :class:`chancy.executors.asyncex.AsyncExecutor`: This executor will execute asyncio
    jobs in the same process as the worker.

You can implement your own executor by subclassing the
:class:`chancy.executors.base.Executor` class and implementing the
:meth:`~chancy.executors.base.Executor.push` method.
"""



import abc
import typing
import traceback
import dataclasses
from datetime import datetime, timezone

from chancy.job import JobInstance
from chancy.queue import Queue

if typing.TYPE_CHECKING:
    from chancy.worker import Worker


class Executor(abc.ABC):
    """
    The base class for all executors.

    Executors are responsible for managing the execution of jobs after they've
    been retrieved from a queue.

    See :class:`~chancy.executors.process.ProcessExecutor` and
    :class:`~chancy.executors.asyncex.AsyncExecutor` for examples of built-in
    executors.
    """

    def __init__(self, worker: "Worker", queue: "Queue"):
        self.worker = worker
        self.queue = queue

    @abc.abstractmethod
    async def push(self, job: JobInstance):
        """
        Push a job onto the job pool.
        """

    async def job_completed(
        self, job: JobInstance, exc: Exception | None = None
    ):
        """
        Called when a job has completed.

        This method should be called by the executor when a job has completed
        execution. It will update the job's state in the queue and handle
        retries if necessary.

        :param job: The job that has completed.
        :param exc: The exception that was raised during execution, if any.
        """
        if exc is not None:
            new_state = (
                JobInstance.State.FAILED
                if job.attempts + 1 >= job.max_attempts
                else JobInstance.State.RETRYING
            )

            new_instance = dataclasses.replace(
                job,
                state=new_state,
                attempts=job.attempts + 1,
                errors=[
                    *job.errors,
                    {
                        "traceback": "".join(
                            traceback.format_exception(
                                type(exc), exc, exc.__traceback__
                            )
                        ),
                        "attempt": job.attempts,
                    },
                ],
            )
        else:
            now = datetime.now(tz=timezone.utc)
            new_instance = dataclasses.replace(
                job,
                state=JobInstance.State.SUCCEEDED,
                completed_at=now,
            )

        await self.worker.queue_update(new_instance)

    @abc.abstractmethod
    def __len__(self):
        """
        Get the number of pending and running jobs in the pool.
        """



import asyncio

from chancy.executors.base import Executor
from chancy.job import JobInstance, Limit
from chancy.utils import import_string


class AsyncExecutor(Executor):
    """
    An Executor which uses asyncio to run its jobs in the main event loop.

    This executor is useful for running large numbers of IO-bound jobs, as it
    can run many jobs concurrently without blocking the main event loop and
    without the high overhead of new processes or threads. However, it is not
    suitable for CPU-bound jobs, as it will block the main event loop and
    prevent other jobs & queues from running.
    """

    def __init__(self, worker, queue):
        super().__init__(worker, queue)
        self.running_jobs: set[asyncio.Task] = set()

    async def push(self, job: JobInstance):
        task = asyncio.create_task(self._job_wrapper(job))
        task.add_done_callback(self._job_cleanup)
        self.running_jobs.add(task)

    def __len__(self):
        return len(self.running_jobs)

    def _job_cleanup(self, task: asyncio.Task):
        self.running_jobs.discard(task)
        task.exception()

    async def _job_wrapper(self, job: JobInstance):
        try:
            func = import_string(job.func)
            if not asyncio.iscoroutinefunction(func):
                raise ValueError(
                    f"Function {job.func} is not a coroutine function"
                )

            kwargs = job.kwargs or {}

            timeout = next(
                (
                    limit.value
                    for limit in job.limits
                    if limit.type_ == Limit.Type.TIME
                ),
                None,
            )

            try:
                # This annoyingly creates quite the excessive traceback,
                # we should revisit this and clean it up.
                await asyncio.wait_for(func(**kwargs), timeout=timeout)
            except (asyncio.TimeoutError, TimeoutError):
                raise asyncio.TimeoutError(
                    f"Job {job.id} timed out after {timeout} seconds"
                )
            self.worker.manager.add(self.job_completed(job))
        except Exception as exc:
            self.worker.manager.add(self.job_completed(job, exc))
